TRANSITION = "transition"
CLONE = "clone"
HALT = "halt"
FINAL = "final"
ACTION = "action"

SYNC_REQ = "sync_req"
SYNC_RES = "sync_res"

NFA_NAME = "nfa_name"
RUN_ID = "run_id"
STATE_FROM = "state_from"
STATE_TO = "state_to"
EVENT = "event"
HISTORY = "history"
HANDLERS = "handlers"
